/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 

*/


#define _WIN32_WINNT 0x0400

#include "windows.h"

#include <iostream>
#include <iomanip>

#define wchar wchar_t
#define int64	__int64

typedef enum
{
	fcRawBytes,
	fcKilo,
	fcMega,
	fcGiga,
	fcTera
}
FormatCode;

FormatCode gFormat = fcRawBytes;

inline void FillInt64(int64 num, wchar* wsz)
{
	_ui64tow (num, wsz, 10);
}

inline Fill2Decimal(unsigned f, wchar *wszIn)
{
	wchar *pwsz = &wszIn[::wcslen(wszIn)];

	*pwsz++ = L'.';
	*pwsz++ = L'0' + f/10;
	*pwsz++ = L'0' + f%10;
	*pwsz = 0;
}

void MyFormatNumber(int64 num, wchar *wszOut, unsigned size)
{
	wchar wszIn[150];

	switch (gFormat)
	{
		case fcKilo:
		{
			static NUMBERFMTW nfmt = {0,0,3, L".", L",", 2};

			FillInt64(num>>10, wszIn);
			::GetNumberFormatW(LOCALE_USER_DEFAULT, 0, wszIn, &nfmt, wszOut, size);
			::wcscat(wszOut, L"K");
		}
		break;

		case fcMega:
		{
			static NUMBERFMTW nfmt = {0,1,3, L".", L",", 2};

			FillInt64(num>>20, wszIn);
			//unsigned f = (num%(1<<20) / ((1<<20)/100));

			//Fill2Decimal(f, wszIn);
			::GetNumberFormatW(LOCALE_USER_DEFAULT, 0, wszIn, &nfmt, wszOut, size);
			::wcscat(wszOut, L"M");
		}
		break;

		case fcGiga:
		{
			static NUMBERFMTW nfmt = {2,1,3, L".", L",", 2};

			FillInt64(num>>30, wszIn);
			unsigned f = (num%(1<<30) / ((1<<30)/100));

			Fill2Decimal(f, wszIn);
			::GetNumberFormatW(LOCALE_USER_DEFAULT, 0, wszIn, &nfmt, wszOut, size);
			::wcscat(wszOut, L"G");
		}
		break;

		case fcTera:
		{
			static NUMBERFMTW nfmt = {2,1,3, L".", L",", 2};

			FillInt64(num>>40, wszIn);
			unsigned f = (num%(int64(1)<<40) / ((int64(1)<<40)/100));

			::GetNumberFormatW(LOCALE_USER_DEFAULT, 0, wszIn, &nfmt, wszOut, size);
			::wcscat(wszOut, L"T");
		}
		break;

		case fcRawBytes:
		default:
		{
			static NUMBERFMTW nfmt = {0,0,3, L".", L",", 2};

			FillInt64(num, wszIn);
			::GetNumberFormatW(LOCALE_USER_DEFAULT, 0, wszIn, &nfmt, wszOut, size);
		}
		break;
	}
}

std::wostream& operator<<(std::wostream& s, int64 num)
{
	wchar wszF[150];
	MyFormatNumber(num, wszF, sizeof(wszF)/sizeof(wchar));

	s << wszF;

	return s;
}

void OutputRow(const wchar *wszTotalSize, const wchar *wszSub, const wchar *wszSize, const wchar *wszFiles, const wchar *wszFileName)
{
	std::wcout
			<< std::setw(20) << wszTotalSize 
			<< std::setw(20) << wszSub 
			<< std::setw(20) << wszSize 
			<< std::setw(10) << wszFiles 
			<< L" " 
			<< wszFileName << std::endl;
}

void DirSize(const wchar *wszFile, int nPrintLevel, int64& totalsize, unsigned& totalfiles, unsigned& totalfolders)
{
	WIN32_FIND_DATAW ffd;
	HANDLE hFind;
	
	unsigned filecount = 0;
	unsigned subfolderfilecount = 0;
	unsigned subfoldercount = 0;
	unsigned foldercount = 0;
	int64 size = 0;
	int64 subfoldersize = 0;

	wchar wszF[MAX_PATH];

	::wcscpy(wszF, wszFile);
	::wcscat(wszF, L"\\*");
	hFind = ::FindFirstFileW(wszF, &ffd);

	if (hFind == INVALID_HANDLE_VALUE)
	{
		std::wcout << L"File not found: "<< wszF << std::endl;
	}
	else
	{
		do
		{
			if(L'.' != ffd.cFileName[0])
			{
				if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					::wcscpy(wszF, wszFile);
					::wcscat(wszF, L"\\");
					::wcscat(wszF, ffd.cFileName);

					DirSize(wszF, nPrintLevel-1, subfoldersize, subfolderfilecount, subfoldercount);
					++foldercount;
				}
				else
				{
					size += int64(ffd.nFileSizeHigh)<<32 | int64(ffd.nFileSizeLow);
					++filecount;
				}
			}
		}
		while(::FindNextFileW(hFind, &ffd));


		totalsize += subfoldersize + size;
		totalfiles += filecount + subfolderfilecount;
		totalfolders += foldercount + subfoldercount;

		if (nPrintLevel>0)
		{
			wchar wsz[100];
			wchar wszSub[150];
			wchar wszFiles[150];
			wchar wszTotal[150];

			MyFormatNumber(size, wszFiles, sizeof(wszFiles)/sizeof(wchar));
			MyFormatNumber(subfoldersize, wszSub, sizeof(wszSub)/sizeof(wchar));
			MyFormatNumber(subfoldersize + size, wszTotal, sizeof(wszTotal)/sizeof(wchar));
			_itow(filecount, wsz, 10);

			OutputRow(wszTotal, wszSub, wszFiles, wsz , wszFile);
		}

		::FindClose(hFind);
	}
}

int _cdecl
wmain(int argc, wchar *argv[])
{
	wchar wszDir[MAX_PATH];
	int displaylevel = MAX_PATH*100;

	::GetCurrentDirectoryW(sizeof(wszDir)/sizeof(wchar), wszDir);

	for (int i = 1; i < argc; i++)
	{
		wchar *pwsz = argv[i];

		switch(pwsz[0])
		{
			case L'/':
			case L'-':
			switch(::towlower(pwsz[1]))
			{
				case L's':
				{
					if (pwsz[2])
						displaylevel = ::_wtoi(&pwsz[2]);
					else
						displaylevel=1;

				}
				break;

				case L'f':
					switch(::towlower(pwsz[2]))
					{
						case 'k' : ::gFormat = fcKilo; break;
						case 'm' : ::gFormat = fcMega; break;
						case 'g' : ::gFormat = fcGiga; break;
						case 't' : ::gFormat = fcTera; break;
						default: 
							::wcscpy(wszDir, pwsz);
						break;
					}
				break;

				case L'?':
					std::wcout << L"usage:\tadu [/s[displaylevel=1]] [folder]" << std::endl;
					return (0);
				break;

				default:
					::wcscpy(wszDir, pwsz);
				break;
			}
			break;

			default:
				::wcscpy(wszDir, pwsz);
			break;
		}
	}

	OutputRow(L"Total", L"Subdirectory", L"Files Size", L"Files" , L"Directory");

	int64 size = 0;
	unsigned files = 0;
	unsigned folders = 0;
	DirSize(wszDir, displaylevel, size, files, folders);

	wchar wsz[100];
	wchar wszTotal[150];

	MyFormatNumber(size, wszTotal, sizeof(wszTotal)/sizeof(wchar));
	_itow(files, wsz, 10);
	OutputRow(wszTotal, L"", L"",  wsz , L"Total");

	int64 diskavailuser, disktotalbytes, disktotalfree;
	::GetDiskFreeSpaceExW(wszDir, (PULARGE_INTEGER)&diskavailuser, (PULARGE_INTEGER)&disktotalbytes, (PULARGE_INTEGER)&disktotalfree);

	std::wcout << std::setw(25) << folders << L" folder(s)" << std::endl <<std::endl;

	std::wcout
		<< diskavailuser << L" bytes available."
		<< L" Total: " << disktotalbytes << L" bytes."
		<< L" " << unsigned ((disktotalfree*100)/disktotalbytes) << L"% free";

	return (0);
}
